package sample.parser;

import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.lightysoft.logmx.business.ParsedEntry;
import com.lightysoft.logmx.mgr.LogFileParser;

/**
 * Sample LogMX Parser able to parse a log file with multi-line support and Relative Date support.<BR/>
 * Here is an example of log file suitable for this parser:<BR/>
 * 

[10/01/2020 12:57:46.718](15328){dbg}{database} In DbServer\DatabaseOperations.cpp:353 (DatabaseOperations::initializeCommonDatabase)
[#1] (B1CB)                                     New common database created, signature is {95F6754B-F809-4854-A847-

[10/01/2020 12:57:46.733](23592){err}{mapi-interface} In StoreProvider\MSProviderImpl.cpp:452 (MSProviderImpl::Logon)
[#1] (common)                                         Exception of class HResultException: StoreProvider\MSProviderImpl.cpp(371), MSProviderImpl::Logon:
                                                      0x8004011c MAPI_E_UNCONFIGURED

[10/01/2020 12:57:50.546](15328){err}{communication} In SCProvider\NtlmAuthenticator.cpp:111 (PocoNtlmAuth::sendResponseGetChallenge)
[#2] (common)                                        NTLM authentication has been unsuccessful

[11/Mar/2020 12:08:24](14668) {https} Task 9032 handler BEGIN 
[28/01/2020 13:58:02.185](21392){err}{communication} In SCProvider\Communicator.cpp:291 (Communicator::checkAndLogResult)
[#124] (B1CB)                                        Exception of class HResultException: SCProvider\HttpConnection.cpp(146), HttpConnection::checkStatus:
                                                     0x80042013 KOFF_E_REQUEST_FAILED (Request info: Batch move message store=aberger@eanw.nexter 
                                                     
 */
public class KerioConnectParser extends LogFileParser {
    /** Current parsed log entry */
    private ParsedEntry entry = null;

    /** Entry date formatter for short numerical dates */
    private SimpleDateFormat dateFormat = null;

    /** Entry date formatter for long dates (month written in English) */
    private SimpleDateFormat dateFormat2 = null;

    /** Entry date formats strings */
    private final String DATE_FORMAT_STRING = "dd/MM/yyyy HH:mm:ss.SSS"; // TODO: is it really "Day/Month" or "Month/Day"?

    private final String DATE_FORMAT2_STRING = "dd/MMM/yyyy HH:mm:ss"; // TODO: is it using English for the month? TODO: make sure the the used abbreviations for the Month are compatible with this Java formatter

    /** Mutex to avoid that multiple threads use the same Date formatter at the same time */
    private final Object DATE_FORMATTER_MUTEX = new Object();

    /** Pattern for entry begin */
    private final static Pattern ENTRY_BEGIN_PATTERN = Pattern.compile(
            "^\\[(\\d{2}/[0-9A-Za-z]{2,}/\\d{4} .*?)\\]\\s*?" // Date
                + "\\((.*?)\\)\\s*?" // Process ID
                + "(?:"//
                + "\\{(.*?)\\}\\s*?" // Level (optional)
                + ")?" //
                + "\\{(.*?)\\}\\s*" // Emitter
                + "(In)?\\s*(.*?)" // Location / Message
                + "$");

    /** Pattern for "Error Type+Code" */
    private final static Pattern ERROR_TYPE_AND_CODE_PATTERN = Pattern.compile(
            "^\\s*?(0x\\S+?)\\s+?(\\S+)(?:\\s+(\\S.*))?$");

    /** Pattern for a blank line */
    private final static Pattern BLANK_LINE_PATTERN = Pattern.compile("^\\s*$");

    /** Pattern for a multiple spaces */
    private final static Pattern SPACES_PATTERN = Pattern.compile("\\s{2,}");

    /** Buffer for Entry message (improves performance for multi-lines entries)  */
    private StringBuilder entryMsgBuffer = null;

    /** Key of user-defined field "Process ID" */
    private static final String CUSTOM_FIELD_PID = "PID";

    /** Key of user-defined field "Location" */
    private static final String CUSTOM_FIELD_LOCATION = "Location";

    /** Key of user-defined field "Error Code" */
    private static final String CUSTOM_FIELD_ERR_CODE = "ErrCode";

    /** Key of user-defined field "Error Type" */
    private static final String CUSTOM_FIELD_ERR_TYPE = "ErrType";

    /** User-defined fields names */
    private static final List<String> EXTRA_FIELDS_KEYS = Arrays.asList(CUSTOM_FIELD_PID,
            CUSTOM_FIELD_LOCATION, CUSTOM_FIELD_ERR_CODE, CUSTOM_FIELD_ERR_TYPE);


    /** 
     * Returns the name of this parser
     * @see com.lightysoft.logmx.mgr.LogFileParser#getParserName()
     */
    @Override
    public String getParserName() {
        return "Kerio Connect Parser";
    }

    /**
     * Returns the supported file type for this parser
     * @see com.lightysoft.logmx.mgr.LogFileParser#getSupportedFileType()
     */
    @Override
    public String getSupportedFileType() {
        return "Kerio Connect log files";
    }

    /**
     * Process the new line of text read from file 
     * @see com.lightysoft.logmx.mgr.LogFileParser#parseLine(java.lang.String)
     */
    @Override
    protected void parseLine(String line) throws Exception {
        // If end of file, records last entry if necessary, and exits
        if (line == null) {
            recordPreviousEntryIfExists();
            return;
        }

        Matcher matcher = ENTRY_BEGIN_PATTERN.matcher(line);
        if (matcher.matches()) {
            // Record previous found entry if exists, then create a new one
            prepareNewEntry();

            entry.setDate(matcher.group(1));
            entry.getUserDefinedFields().put(CUSTOM_FIELD_PID, matcher.group(2));
            entry.setLevel(matcher.group(3));
            entry.setEmitter(matcher.group(4));
            if (matcher.group(5) == null) {
                entryMsgBuffer.append(matcher.group(6)); // No "In" word => add the string to the entry message
            } else {
                entry.getUserDefinedFields().put(CUSTOM_FIELD_LOCATION, matcher.group(6)); // "In" word => Location
            }
        } else if (entry != null) {
            matcher = ERROR_TYPE_AND_CODE_PATTERN.matcher(line);
            if (matcher.matches()) {
                entry.getUserDefinedFields().put(CUSTOM_FIELD_ERR_CODE, matcher.group(1));
                entry.getUserDefinedFields().put(CUSTOM_FIELD_ERR_TYPE, matcher.group(2));
                String msg = matcher.group(3);
                if (msg != null) {
                    appendLineToMessage(msg);
                }
            } else if (!BLANK_LINE_PATTERN.matcher(line).matches()) {
                // appends this line to previous entry message
                appendLineToMessage(SPACES_PATTERN.matcher(line).replaceAll(" ")); // remove multiple spaces
            }
        }
    }

    /** 
     * Returns the ordered list of user-defined fields to display (given by their key), for each entry.
     * @see com.lightysoft.logmx.mgr.LogFileParser#getUserDefinedFields()
     */
    @Override
    public List<String> getUserDefinedFields() {
        return EXTRA_FIELDS_KEYS;
    }

    /**
     * Returns a relative Date for the given entry 
     * @see com.lightysoft.logmx.mgr.LogFileParser#getRelativeEntryDate(com.lightysoft.logmx.business.ParsedEntry)
     */
    @Override
    public Date getRelativeEntryDate(ParsedEntry pEntry) throws Exception {
        return getAbsoluteEntryDate(pEntry);
    }

    /**
     * Returns the absolute Date for the given entry 
     * @see com.lightysoft.logmx.mgr.LogFileParser#getAbsoluteEntryDate(com.lightysoft.logmx.business.ParsedEntry)
     */
    @Override
    public Date getAbsoluteEntryDate(ParsedEntry pEntry) throws Exception {
        synchronized (DATE_FORMATTER_MUTEX) { // Java date formatter is not thread-safe
            if (dateFormat == null) {
                // Now create the date formatter using the right Locale
                // (method "getLocale()" can't be called from the constructor) 
                dateFormat = new SimpleDateFormat(DATE_FORMAT_STRING, getLocale());
                dateFormat2 = new SimpleDateFormat(DATE_FORMAT2_STRING, getLocale());
            }
            try {
                return dateFormat.parse(pEntry.getDate());
            } catch (Exception e) {
                return dateFormat2.parse(pEntry.getDate());
            }
        }
    }

    /**
     * Send to LogMX the current parsed log entry
     * @throws Exception
     */
    private void recordPreviousEntryIfExists() throws Exception {
        if (entry != null) {
            entry.setMessage(entryMsgBuffer.toString());
            addEntry(entry);
        }
    }

    /**
     * Send to LogMX the current parsed log entry, then create a new one
     * @throws Exception
     */
    private void prepareNewEntry() throws Exception {
        recordPreviousEntryIfExists();
        entry = createNewEntry();
        entryMsgBuffer = new StringBuilder(100);
        entry.setUserDefinedFields(new HashMap<String, Object>(4));
    }

    /**
     * Add the given string to the current entry message, by prepending a '\n' first if the message already contains some text
     * @param str Line to add to the entry message 
     */
    private void appendLineToMessage(String str) {
        if (entryMsgBuffer.length() != 0) {
            entryMsgBuffer.append('\n');
        }
        entryMsgBuffer.append(str);
    }
}
